/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: automation_line.h,v 1.19 2004/01/15 01:56:10 pbd Exp $
*/

#ifndef __ardour_automation_line_h__
#define __ardour_automation_line_h__

#include <vector>
#include <string>
#include <sys/types.h>
#include <gtk--.h>
#include <gtk-canvas.h>
#include <sigc++/signal_system.h>

#include <pbd/undo.h>

#include <ardour/automation_event.h>

class PublicEditor;
class AutomationLine;

namespace ARDOUR {
	class AutomationList;
	class Curve;
}

class ControlPoint 
{
  public:
	ControlPoint (AutomationLine& al, gint (*event_handler)(GtkCanvasItem*, GdkEvent*, gpointer));
	ControlPoint (const ControlPoint&, bool dummy_arg_to_force_special_copy_constructor);
	~ControlPoint ();

	enum ShapeType {
		Full,
		Start,
		End
	};
	
	void move_to (double x, double y, ShapeType);
	void reset (double x, double y, unsigned long, unsigned long, ShapeType);
	double get_x() const { return _x; }
	double get_y() const { return _y; }

	void hide (); 
	void show ();
	
	void set_size (double);
	
	GtkCanvasItem* item;
	AutomationLine& line;
	unsigned long view_index;
	unsigned long model_index;
	bool valid;
	bool can_slide;

  private:
	double _x;
	double _y;
	double _size;
	ShapeType _shape;
};

class AutomationLine : public SigC::Object
{
  public:
	AutomationLine (string name, PublicEditor&, GtkCanvasItem *,
			gint (*point_event_handler)(GtkCanvasItem*, GdkEvent*, gpointer),
			gint (*line_event_handler)(GtkCanvasItem*, GdkEvent*, gpointer));

	virtual ~AutomationLine ();

	virtual void reset () = 0;

	virtual void clear() = 0;

	void remove_point (ControlPoint&);
	bool control_points_adjacent (double xval, unsigned long& before, unsigned long& after);

	/* dragging API */

	virtual void start_drag (ControlPoint*, float fraction);
	virtual void point_drag(ControlPoint&, jack_nframes_t x, float, bool with_push);
	virtual void end_drag (ControlPoint*);
	virtual void line_drag(unsigned long i1, unsigned long i2, float, bool with_push);

	ControlPoint* nth (unsigned int);

	string  name() const { return _name; }
	bool    visible() const { return _visible; }
	guint32 height() const { return _height; }

	void         set_line_color (unsigned int);
	unsigned int get_line_color() const { return _line_color; }

	void    show ();
	void    hide ();
	void    set_height (guint32);
	PublicEditor& editor() const { return _editor; }
	void    set_verbose_cursor_uses_gain_mapping (bool yn);

	void queue_reset ();

	virtual string  get_verbose_cursor_string (float);

	GtkCanvasGroup* canvas_group() const { return GTK_CANVAS_GROUP(group); }
	GtkCanvasItem*  parent_group() const { return _parent_group; }
	GtkCanvasItem*  grab_item() const { return line; }

	void show_selection ();
	void hide_selection ();

	void set_point_size (double size);

	static void invalidate_point (GtkCanvasPoints*, unsigned long index);
	static bool invalid_point (GtkCanvasPoints*, unsigned long index);
	
	virtual void view_to_model_y (double&) = 0;
	virtual void model_to_view_y (double&) = 0;

  protected:
	PublicEditor& _editor;
	guint32 _height;
	unsigned int _line_color;

	bool    _visible  : 1;
	bool    _vc_uses_gain_mapping : 1;
	bool    terminal_points_can_slide : 1;
	bool    update_pending : 1;
	bool    no_draw : 1;

	string  _name;

	GtkCanvasItem*  _parent_group;
	GtkCanvasItem*   group;
	GtkCanvasItem*   line; /* line */
	GtkCanvasPoints* point_coords; /* coordinates for canvas line */
	vector<ControlPoint*> control_points; /* visible control points */

	gint   (*point_callback)(GtkCanvasItem*, GdkEvent*, gpointer);

	void hide_and_invalidate ();

	void determine_visible_control_points (GtkCanvasPoints*);
	void sync_model_from (ControlPoint&);
	void sync_model_with_view_point (ControlPoint&);
	void sync_model_with_view_line (ControlPoint&, ControlPoint&);
	void modify_view (ControlPoint&, double, double, bool with_push);
	
	virtual void change_model (unsigned long, double x, double y) = 0;
	virtual void change_model_range (unsigned long, unsigned long, double) = 0;
	virtual bool model_nth (unsigned long, double&, double&) = 0;
	virtual void remove_model_range (unsigned long start, unsigned long end) = 0;

	virtual UndoAction get_memento() = 0;
	
  private:
	unsigned long drags;
	double   first_drag_fraction;
	double   last_drag_fraction;
	unsigned long line_drag_cp1;
	unsigned long line_drag_cp2;

	void modify_view_point(ControlPoint&, double, double, bool with_push);
	void reset_line_coords (ControlPoint&);
	void update_line ();

	struct ModelRepresentation {
	    unsigned long start;
	    unsigned long end;
	    double xpos;
	    double ypos;
	    double xmin;
	    double ymin;
	    double xmax;
	    double ymax;
	    double xval;
	    double yval;
	};

	void model_representation (ControlPoint&, ModelRepresentation&);
};

class AutomationLineList : public AutomationLine
{
  public:
	AutomationLineList (string name, PublicEditor&, GtkCanvasItem*, ARDOUR::AutomationList&,
			    gint (*point_callback)(GtkCanvasItem*, GdkEvent*, gpointer),
			    gint (*line_callback)(GtkCanvasItem*, GdkEvent*, gpointer));

	~AutomationLineList ();

	void reset ();
	void clear ();

  protected:
	ARDOUR::AutomationList& alist;
	virtual void reset_callback (const ARDOUR::AutomationList&);
	void list_changed ();

	bool model_nth (unsigned long, double&, double&);
};

class AutomationLineCurve : public AutomationLine
{
  public:
	AutomationLineCurve (string name, PublicEditor&, GtkCanvasItem*, ARDOUR::Curve&,
			     gint (*point_callback)(GtkCanvasItem*, GdkEvent*, gpointer),
			     gint (*line_callback)(GtkCanvasItem*, GdkEvent*, gpointer));

	~AutomationLineCurve ();

	void reset ();
	void clear ();

	UndoAction get_memento();

  private:
	ARDOUR::Curve& curve;

	void reset_callback (size_t, double *, double *);
	void change_model (unsigned long, double x, double y);
	void change_model_range (unsigned long, unsigned long, double delta);
	void remove_model_range (unsigned long start, unsigned long end);
	bool model_nth (unsigned long, double&, double&);
};	


#endif /* __ardour_automation_line_h__ */

