#include <stdio.h>
#include <math.h>
#include <gtk-canvas.h>

#include "canvas-curve.h"
#include "rgb_macros.h"

enum {
	ARG_0,
	ARG_X1,
	ARG_Y1,
	ARG_X2,
	ARG_Y2,
	ARG_CURVE,
	ARG_COLOR_RGBA
};

static void gtk_canvas_curve_class_init (GtkCanvasCurveClass *class);
static void gtk_canvas_curve_init       (GtkCanvasCurve      *curve);
static void gtk_canvas_curve_set_arg    (GtkObject              *object,
					 GtkArg                 *arg,
					 guint                   arg_id);
static void gtk_canvas_curve_get_arg    (GtkObject              *object,
					 GtkArg                 *arg,
					 guint                   arg_id);

static void   gtk_canvas_curve_update      (GtkCanvasItem *item, double *affine, ArtSVP *clip_path, int flags);
static void   gtk_canvas_curve_bounds      (GtkCanvasItem *item, double *x1, double *y1, double *x2, double *y2);
static double gtk_canvas_curve_point (GtkCanvasItem *item, double x, double y, int cx, int cy, GtkCanvasItem **actual_item);
static void   gtk_canvas_curve_render (GtkCanvasItem *item, GtkCanvasBuf *buf);
static void   gtk_canvas_curve_draw (GtkCanvasItem *item, GdkDrawable *drawable, int x, int y, int w, int h);

static GtkCanvasItemClass *parent_class;

extern curve_get_vector (double, double, float*, size_t, void*);

GtkType
gtk_canvas_curve_get_type (void)
{
	static GtkType curve_type = 0;

	if (!curve_type) {
		GtkTypeInfo curve_info = {
			"GtkCanvasCurve",
			sizeof (GtkCanvasCurve),
			sizeof (GtkCanvasCurveClass),
			(GtkClassInitFunc) gtk_canvas_curve_class_init,
			(GtkObjectInitFunc) gtk_canvas_curve_init,
			NULL, /* reserved_1 */
			NULL, /* reserved_2 */
			(GtkClassInitFunc) NULL
		};

		curve_type = gtk_type_unique (gtk_canvas_item_get_type (), &curve_info);
	}

	return curve_type;
}

static void
gtk_canvas_curve_class_init (GtkCanvasCurveClass *class)
{
	GtkObjectClass *object_class;
	GtkCanvasItemClass *item_class;

	object_class = (GtkObjectClass *) class;
	item_class = (GtkCanvasItemClass *) class;

	parent_class = gtk_type_class (gtk_canvas_item_get_type ());

	gtk_object_add_arg_type ("GtkCanvasCurve::x1", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_X1);
	gtk_object_add_arg_type ("GtkCanvasCurve::y1", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_Y1);
	gtk_object_add_arg_type ("GtkCanvasCurve::x2", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_X2);
	gtk_object_add_arg_type ("GtkCanvasCurve::y2", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_Y2);
	gtk_object_add_arg_type ("GtkCanvasCurve::curve", GTK_TYPE_POINTER, GTK_ARG_READWRITE, ARG_CURVE);
	gtk_object_add_arg_type ("GtkCanvasCurve::color_rgba", GTK_TYPE_INT, GTK_ARG_READWRITE, ARG_COLOR_RGBA);

	object_class->set_arg = gtk_canvas_curve_set_arg;
	object_class->get_arg = gtk_canvas_curve_get_arg;

	item_class->update = gtk_canvas_curve_update;
	item_class->bounds = gtk_canvas_curve_bounds;
	item_class->point = gtk_canvas_curve_point;
	item_class->render = gtk_canvas_curve_render;
	item_class->draw = gtk_canvas_curve_draw;
}

static void
gtk_canvas_curve_init (GtkCanvasCurve *curve)
{
	curve->x1 = 0.0;
	curve->y1 = 0.0;
	curve->x2 = 0.0;
	curve->y2 = 0.0;
	curve->curve_arg = 0;
	curve->vector = 0;
	curve->veclen = 0;
	curve->color = RGBA_TO_UINT(98,123,174,241);
}

static void
gtk_canvas_curve_bounds_world (GtkCanvasItem *item, int* ix1, int* iy1, int* ix2, int* iy2)
{
	double x1, x2, y1, y2;
	ArtPoint i1, i2;
	ArtPoint w1, w2;
	double i2w[6];

	gtk_canvas_curve_bounds (item, &x1, &y1, &x2, &y2);

	i1.x = x1;
	i1.y = y1;
	i2.x = x2;
	i2.y = y2;
	
	gtk_canvas_item_i2w_affine (item, i2w);
	art_affine_point (&w1, &i1, i2w);
	art_affine_point (&w2, &i2, i2w);

	*ix1 = (int) rint(w1.x);
	*ix2 = (int) rint(w2.x);
	*iy1 = (int) rint(w1.y);
	*iy2 = (int) rint(w2.y);

	*iy2 += 1;
}

static void 
gtk_canvas_curve_reset_bounds (GtkCanvasItem *item)
{
	int Ix1, Ix2, Iy1, Iy2;

	gtk_canvas_curve_bounds_world (item, &Ix1, &Iy1, &Ix2, &Iy2);
	gtk_canvas_update_bbox (item, Ix1, Iy1, Ix2, Iy2);
}

/* 
 * CANVAS CALLBACKS 
 */

static void
gtk_canvas_curve_set_arg (GtkObject *object, GtkArg *arg, guint arg_id)
{
	GtkCanvasItem *item;
	GtkCanvasCurve *curve;
	int redraw;
	int calc_bounds;

	item = GTK_CANVAS_ITEM (object);
	curve = GTK_CANVAS_CURVE (object);

	redraw = FALSE;
	calc_bounds = FALSE;

	switch (arg_id) {
	case ARG_X1:
	        if (curve->x1 != GTK_VALUE_DOUBLE (*arg)) {
		        curve->x1 = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_Y1:
	        if (curve->y1 != GTK_VALUE_DOUBLE (*arg)) {
		        curve->y1 = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_X2:
	        if (curve->x2 != GTK_VALUE_DOUBLE (*arg)) {
		        curve->x2 = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_Y2:
	        if (curve->y2 != GTK_VALUE_DOUBLE (*arg)) {
		        curve->y2 = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_CURVE:
	        if (curve->curve_arg != GTK_VALUE_POINTER (*arg)) {
		        curve->curve_arg = GTK_VALUE_POINTER (*arg);
			redraw = TRUE;
		}
		break;

	case ARG_COLOR_RGBA:
		if (curve->color != GTK_VALUE_INT(*arg)) {
			curve->color = GTK_VALUE_INT(*arg);
			UINT_TO_RGBA (curve->color, &curve->r, &curve->g, &curve->b, &curve->a);
			redraw = TRUE;
		}
		break;

	default:
		break;
	}

	if (calc_bounds) {

		gtk_canvas_item_request_update (item);

	} else if (redraw) {

		int Ix1, Ix2, Iy1, Iy2;
		gtk_canvas_curve_bounds_world (item, &Ix1, &Iy1, &Ix2, &Iy2);
		gtk_canvas_request_redraw (item->canvas, Ix1, Iy1, Ix2, Iy2);
	}
}

static void
gtk_canvas_curve_get_arg (GtkObject *object, GtkArg *arg, guint arg_id)
{
	GtkCanvasCurve *curve;

	curve = GTK_CANVAS_CURVE (object);

	switch (arg_id) {
	case ARG_X1:
		GTK_VALUE_DOUBLE (*arg) = curve->x1;
		break;
	case ARG_Y1:
		GTK_VALUE_DOUBLE (*arg) = curve->y1;
		break;
	case ARG_X2:
		GTK_VALUE_DOUBLE (*arg) = curve->x2;
		break;
	case ARG_Y2:
		GTK_VALUE_DOUBLE (*arg) = curve->y2;
		break;
	case ARG_CURVE:
		GTK_VALUE_POINTER (*arg) = curve->curve_arg;
		break;
	case ARG_COLOR_RGBA:
		GTK_VALUE_INT (*arg) = curve->color;
		break;
	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

static void
gtk_canvas_curve_update (GtkCanvasItem *item, double *affine, ArtSVP *clip_path, int flags)
{
	GtkCanvasCurve *curve;
	double x;
	double y;
	int length;

	curve = GTK_CANVAS_CURVE (item);

	if (parent_class->update)
		(* parent_class->update) (item, affine, clip_path, flags);

	gtk_canvas_curve_reset_bounds (item);

	x = curve->x1;
	y = curve->y1;

	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c (GTK_CANVAS(item->canvas), x, y, &curve->bbox_ulx, &curve->bbox_uly);

	x = curve->x2;
	y = curve->y2;

	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c (GTK_CANVAS(item->canvas), x, y, &curve->bbox_lrx, &curve->bbox_lry);

	length = (int) floor (curve->x2 - curve->x1);
	
	if (curve->veclen < length) {
		if (curve->vector) {
			free (curve->vector);
		}
		curve->vector = (float*) malloc (sizeof (float) * length);
		curve->veclen = length;
	}

	curve_get_vector (0.0, -1.0, curve->vector, length, curve->curve_arg);
}

static void
gtk_canvas_curve_render (GtkCanvasItem *item,
			      GtkCanvasBuf *buf)
{
	GtkCanvasCurve *curve;
	int end, begin;
	float height;
	int len;
	int n;
	int offset;
	int y;

	curve = GTK_CANVAS_CURVE (item);

	if (parent_class->render) {
		(*parent_class->render) (item, buf);
	}

	if (buf->is_bg) {
		gtk_canvas_buf_ensure_buf (buf);
		buf->is_bg = FALSE;
	}

	begin = MAX(curve->bbox_ulx,buf->rect.x0);
	end = MIN(curve->bbox_lrx,buf->rect.x1);

	offset = begin - curve->bbox_ulx;

	// begin = curve->bbox_ulx;
	// end = curve->bbox_lrx;

	height = (float) (curve->y2 - curve->y1);

	len = end - begin;
	len = MIN((curve->veclen - offset), len);

	y = (int) floor (curve->bbox_uly + ((1.0 - curve->vector[offset]) * height));
	--len;

	for (n = 1; n < len; ++n) {
		int next;
		int t, b;
		
		next = (gint) floor (curve->bbox_uly + ((1.0 - curve->vector[n+offset]) * height));

		if (next < y) {
			t = y;
			b = next;
		} else {
			t = next;
			b = y;
		}

		if (y == next) {
			PAINT_HORIZ (buf, curve->r, curve->g, curve->b, (begin+n), begin+n+1, y);
		} else {
			PAINT_VERT (buf, curve->r, curve->g, curve->b, (begin+n), b, t);
		}

		y = next;
	}
}

static void
gtk_canvas_curve_draw (GtkCanvasItem *item,
		       GdkDrawable *drawable,
		       int x, int y,
		       int width, int height)
{
	GtkCanvasCurve *curve;

	curve = GTK_CANVAS_CURVE (item);

	if (parent_class->draw) {
		(* parent_class->draw) (item, drawable, x, y, width, height);
	}

	fprintf (stderr, "please don't use the CanvasCurve item in a non-aa Canvas\n");
	abort ();
}

static void
gtk_canvas_curve_bounds (GtkCanvasItem *item, double *x1, double *y1, double *x2, double *y2)
{
	GtkCanvasCurve *curve = GTK_CANVAS_CURVE (item);

	*x1 = curve->x1;
	*y1 = curve->y1;
	*x2 = curve->x2;
	*y2 = curve->y2;
}

static double
gtk_canvas_curve_point (GtkCanvasItem *item, double x, double y, int cx, int cy, GtkCanvasItem **actual_item)
{
	GtkCanvasCurve *curve;
	double x1, y1, x2, y2;
	double dx, dy;

	curve = GTK_CANVAS_CURVE (item);

	*actual_item = item;

	/* Find the bounds for the rectangle plus its outline width */

	gtk_canvas_curve_bounds (item, &x1, &y1, &x2, &y2);

	/* Is point inside rectangle */
	
	if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
		return 0.0;
	}

	/* Point is outside rectangle */

	if (x < x1)
		dx = x1 - x;
	else if (x > x2)
		dx = x - x2;
	else
		dx = 0.0;

	if (y < y1)
		dy = y1 - y;
	else if (y > y2)
		dy = y - y2;
	else
		dy = 0.0;

	return sqrt (dx * dx + dy * dy);
}
