/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: editor_markers.cc,v 1.13 2004/01/15 01:56:10 pbd Exp $
*/

#include <stdlib.h>
#include <math.h>

#include <gtk-canvas.h>

#include <ardour/location.h>

#include "editor.h"
#include "marker.h"

#include "i18n.h"

using namespace SigC;
using namespace ARDOUR;
using namespace Gtk;

void
Editor::clear_markers ()
{
	for (LocationMarkerMap::iterator i = location_markers.begin(); i != location_markers.end(); ++i) {
		delete i->second;
	}

	location_markers.clear ();
}

void
Editor::add_new_location (Location *location)
{
	LocationMarkers *lam = new LocationMarkers;
	unsigned int color;

	if (location->is_cd_marker()) {
		color = location_cd_marker_color;
	} else if (location->is_mark()) {
		color = location_marker_color;
	} else {
		color = location_range_color;
	}

	if (location->is_mark()) {
		lam->start = new Marker (*this, GTK_CANVAS_GROUP(marker_group), color, 
					 location->name(), Marker::Mark, PublicEditor::canvas_marker_event, location->start());
		lam->end   = 0;
	} else {
		lam->start = new Marker (*this, GTK_CANVAS_GROUP(marker_group), color, 
					 location->name(), Marker::Start, PublicEditor::canvas_marker_event, location->start());
		lam->end   = new Marker (*this, GTK_CANVAS_GROUP(marker_group), color, 
					 location->name(), Marker::End, PublicEditor::canvas_marker_event, location->end());
	}

	if (location->is_hidden ()) {
		lam->hide();
	} else {
		lam->show ();
	}

	location->start_changed.connect (slot (*this, &Editor::location_changed));
	location->end_changed.connect (slot (*this, &Editor::location_changed));
	location->changed.connect (slot (*this, &Editor::location_changed));
	location->name_changed.connect (slot (*this, &Editor::location_changed));
	location->FlagsChanged.connect (slot (*this, &Editor::location_flags_changed));

	pair<Location*,LocationMarkers*> newpair;

	newpair.first = location;
	newpair.second = lam;

	location_markers.insert (newpair);
}

void
Editor::location_changed (Location *location)
{
	LocationMarkers *lam = find_location_markers (location);

	if (lam == 0) {
		fatal << _("programming error: location/marker map does not contain location!") << endmsg;
		/*NOTREACHED*/
	}
	
	lam->set_name (location->name());
	lam->set_position (location->start(), location->end());
}

void
Editor::location_flags_changed (Location *location, void *src)
{
	LocationMarkers *lam = find_location_markers (location);

	if (lam == 0) {
		fatal << _("programming error: location/marker map does not contain location!") << endmsg;
		/*NOTREACHED*/
	}

	if (location->is_cd_marker()) {
		lam->set_color_rgba (location_cd_marker_color);
	} else if (location->is_mark()) {
		lam->set_color_rgba (location_marker_color);
	} else {
		lam->set_color_rgba (location_range_color);
	}
	
	if (location->is_hidden()) {
		lam->hide();
	} else {
		lam->show ();
	}
}

Editor::LocationMarkers::~LocationMarkers ()
{
	if (start) {
		delete start;
	}

	if (end) {
		delete end;
	}
}

Editor::LocationMarkers *
Editor::find_location_markers (Location *location)
{
	LocationMarkerMap::iterator i;

	for (i = location_markers.begin(); i != location_markers.end(); ++i) {
		if ((*i).first == location) {
			return (*i).second;
		}
	}

	return 0;
}

Location *
Editor::find_location_from_marker (Marker *marker, bool& is_start)
{
	LocationMarkerMap::iterator i;

	for (i = location_markers.begin(); i != location_markers.end(); ++i) {
		LocationMarkers *lm = (*i).second;
		if (lm->start == marker) {
			is_start = true;
			return (*i).first;
		} else if (lm->end == marker) {
			is_start = false;
			return (*i).first;
		}
	}

	return 0;
}

void
Editor::refresh_location_display_internal (Locations::LocationList& locations)
{
	clear_markers ();
	
	for (Locations::LocationList::iterator i = locations.begin(); i != locations.end(); ++i) {
		add_new_location (*i);
	}
}

void
Editor::refresh_location_display ()
{
	if (session) {
		session->locations()->apply (*this, &Editor::refresh_location_display_internal);
	}
}


void
Editor::LocationMarkers::hide() 
{
	start->hide ();
	if (end) { end->hide(); }
}

void
Editor::LocationMarkers::show() 
{
	start->show ();
	if (end) { end->show(); }
}

void
Editor::LocationMarkers::set_name (const string& str) 
{
	start->set_name (str);
	if (end) { end->set_name (str); }
}

void
Editor::LocationMarkers::set_position (jack_nframes_t startf, 
										jack_nframes_t endf) 
{
	start->set_position (startf);
	if (end) { end->set_position (endf); }
}

void
Editor::LocationMarkers::set_color_rgba (unsigned int rgba) 
{
	start->set_color_rgba (rgba);
	if (end) { end->set_color_rgba (rgba); }
}

void
Editor::mouse_add_new_marker (GdkEvent* ev)
{
	if (session) {
		jack_nframes_t where = event_frame (ev);
		Location *location = new Location (where, where, "mark", Location::IsMark);
		session->locations()->add (location, true);
	}
}

void
Editor::remove_marker (GtkCanvasItem* item, GdkEvent* event)
{
	Marker* marker;
	bool is_start;

	if ((marker = static_cast<Marker *> (gtk_object_get_data (GTK_OBJECT(item), "marker"))) == 0) {
		fatal << _("programming error: marker canvas item has no marker object pointer!") << endmsg;
		/*NOTREACHED*/
	}

	Location* loc = find_location_from_marker (marker, is_start);

	if (session && loc) {
		Gtk::Main::idle.connect (bind (slot (*this, &Editor::really_remove_marker), loc));
	}
}

gint
Editor::really_remove_marker (Location* loc)
{
	/* XXX needs to be reversible */

	session->locations()->remove (loc);

	return FALSE;
}

void
Editor::location_gone (Location *location)
{
	LocationMarkerMap::iterator i;

	for (i = location_markers.begin(); i != location_markers.end(); ++i) {
		if ((*i).first == location) {
			delete (*i).second;
			location_markers.erase (i);
			break;
		}
	}
}

void
Editor::tm_marker_context_menu (GdkEventButton* ev, GtkCanvasItem* item)
{
	if (tm_marker_menu == 0) {
		build_tm_marker_menu ();
	}

	marker_menu_item = item;
	tm_marker_menu->popup (1, ev->time);

}


void
Editor::marker_context_menu (GdkEventButton* ev, GtkCanvasItem* item)
{
	if (marker_menu == 0) {
		build_marker_menu ();
	}

	marker_menu_item = item;
	marker_menu->popup (1, ev->time);

}

void
Editor::build_marker_menu ()
{
	using namespace Menu_Helpers;

	marker_menu = new Menu;
	MenuList& items = marker_menu->items();

	items.push_back (MenuElem (_("Rename"), slot (*this, &Editor::marker_menu_rename)));
	items.push_back (MenuElem (_("Remove"), slot (*this, &Editor::marker_menu_remove)));
}

void
Editor::build_tm_marker_menu ()
{
	using namespace Menu_Helpers;

	tm_marker_menu = new Menu;
	MenuList& items = tm_marker_menu->items();

	items.push_back (MenuElem (_("Edit"), slot (*this, &Editor::marker_menu_edit)));
	items.push_back (MenuElem (_("Remove"), slot (*this, &Editor::marker_menu_remove)));
}

void
Editor::marker_menu_edit ()
{
	MeterMarker* mm;
	TempoMarker* tm;
	Marker* marker;

	if ((marker = reinterpret_cast<Marker *> (gtk_object_get_data (GTK_OBJECT(marker_menu_item), "marker"))) == 0) {
		fatal << _("programming error: marker canvas item has no marker object pointer!") << endmsg;
		/*NOTREACHED*/
	}

	if ((mm = dynamic_cast<MeterMarker*> (marker)) != 0) {
		edit_meter_section (&mm->meter());
	} else if ((tm = dynamic_cast<TempoMarker*> (marker)) != 0) {
		edit_tempo_section (&tm->tempo());
	} else {
		fatal << X_("programming erorr: unhandled marker type in Editor::marker_menu_edit")
		      << endmsg;
		/*NOTREACHED*/
	}
}

void
Editor::marker_menu_remove ()
{
	MeterMarker* mm;
	TempoMarker* tm;
	Marker* marker;

	if ((marker = reinterpret_cast<Marker *> (gtk_object_get_data (GTK_OBJECT(marker_menu_item), "marker"))) == 0) {
		fatal << _("programming error: marker canvas item has no marker object pointer!") << endmsg;
		/*NOTREACHED*/
	}

	if ((mm = dynamic_cast<MeterMarker*> (marker)) != 0) {
		remove_meter_marker (marker_menu_item);
	} else if ((tm = dynamic_cast<TempoMarker*> (marker)) != 0) {
		remove_tempo_marker (marker_menu_item);
	} else {
		remove_marker (marker_menu_item, (GdkEvent*) 0);
	}
}

void
Editor::marker_menu_rename ()
{
	Marker* marker;

	if ((marker = reinterpret_cast<Marker *> (gtk_object_get_data (GTK_OBJECT(marker_menu_item), "marker"))) == 0) {
		fatal << _("programming error: marker canvas item has no marker object pointer!") << endmsg;
		/*NOTREACHED*/
	}

	Dialog dialog;
	Entry  entry;
	Button ok_button (_("OK"));
	Button cancel_button (_("Cancel"));

	dialog.set_title (_("ardour: rename mark"));
	dialog.set_name ("MarkRenameWindow");
	dialog.set_usize (300, -1);
	dialog.set_position (GTK_WIN_POS_MOUSE);
	dialog.set_modal (true);

	dialog.get_vbox()->set_border_width (10);
	dialog.get_vbox()->pack_start (entry);
	dialog.get_action_area()->pack_start (ok_button);
	dialog.get_action_area()->pack_start (cancel_button);

	entry.set_name ("MarkerNameDisplay");
	ok_button.set_name ("EditorGTKButton");
	cancel_button.set_name ("EditorGTKButton");

	entry.activate.connect (bind (slot (*this, &Editor::finish_sub_event_loop), 1));
	cancel_button.clicked.connect (bind (slot (*this, &Editor::finish_sub_event_loop), -1));
	ok_button.clicked.connect (bind (slot (*this, &Editor::finish_sub_event_loop), 1));
	dialog.delete_event.connect (bind (slot (*this, &Editor::finish_sub_event_loop_on_delete), -1));

	/* recurse */
	
	dialog.show_all ();

	run_sub_event_loop ();

	if (sub_event_loop_status == 1) {

		Location* l;
		bool is_start;

		if ((l = find_location_from_marker (marker, is_start)) != 0) {
			l->set_name (entry.get_text());
		}
	}
}

