/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: selection.cc,v 1.13 2003/11/12 20:10:14 essej Exp $
*/

#include <algorithm>
#include <sigc++/bind.h>
#include <pbd/error.h>

#include <ardour/route.h>

#include "regionview.h"
#include "selection.h"
#include "selection_templates.h"
#include "time_axis_view.h"

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

struct AudioRangeComparator {
    bool operator()(AudioRange a, AudioRange b) {
	    return a.start < b.start;
    }
};

Selection&
Selection::operator= (const Selection& other)
{
	if (&other != this) {
		audio_regions = other.audio_regions;
		tracks = other.tracks;
		time = other.time;
		lines = other.lines;
	}
	return *this;
}

bool
operator== (const Selection& a, const Selection& b)
{
	return a.audio_regions == b.audio_regions &&
		a.tracks == b.tracks &&
		a.time.track == b.time.track &&
		a.time.group == b.time.group && 
		a.time == b.time &&
		a.lines == b.lines;
}

void
Selection::clear ()
{
	clear_tracks ();
	clear_audio_regions ();
	clear_time ();
	clear_lines ();
}

void
Selection::dump_region_layers()
{
	cerr << "region selection layer dump" << endl;
	for (AudioRegionSelection::iterator i = audio_regions.begin(); i != audio_regions.end(); ++i) {
		cerr << "layer: " << (int)(*i)->region.layer() << endl;
	}
}

void
Selection::clear_audio_regions ()
{
	audio_regions.clear_all ();
	RegionsChanged();
}

void
Selection::clear_tracks ()
{
	tracks.clear ();
	RoutesChanged();
}

void
Selection::clear_time ()
{
	time.track = 0;
	time.group = 0;
	time.clear();

	TimeChanged ();
}

void
Selection::clear_lines ()
{
	lines.clear ();
	LinesChanged();
}

void
Selection::add (TimeAxisView* track)
{
	void (Selection::*pmf)(TimeAxisView*) = &Selection::remove;
	track->GoingAway.connect (bind (slot (*this, pmf), track));
	tracks.push_back (track);
	RoutesChanged();
}

void
Selection::add (AudioRegionView* r)
{
	audio_regions.add (r);
	RegionsChanged ();
}

long
Selection::add (jack_nframes_t start, jack_nframes_t end)
{
	AudioRangeComparator cmp;

	time.push_back (AudioRange (start, end, next_time_id++));
	time.sort (cmp);
	
	TimeChanged ();

	return next_time_id - 1;
}

void
Selection::replace (unsigned long id, jack_nframes_t start, jack_nframes_t end)
{
	for (list<AudioRange>::iterator i = time.begin(); i != time.end(); ++i) {
		if ((*i).id == id) {
			time.erase (i);
			time.push_back (AudioRange(start,end, id));

			AudioRangeComparator cmp;
			time.sort (cmp);

			TimeChanged ();
			break;
		}
	}
}

void
Selection::add (AutomationLine* ac)
{
	lines.push_back (ac);
	LinesChanged();
}

void
Selection::remove (TimeAxisView* track)
{
	list<TimeAxisView*>::iterator i;
	if ((i = find (tracks.begin(), tracks.end(), track)) != tracks.end()) {
		tracks.erase (i);
		RoutesChanged();
	}
}

void
Selection::remove (AudioRegionView* r)
{
	audio_regions.remove (r);
	RegionsChanged ();
}


void
Selection::remove (unsigned long selection_id)
{
	if (time.empty()) {
		return;
	}

	for (list<AudioRange>::iterator i = time.begin(); i != time.end(); ++i) {
		if ((*i).id == selection_id) {
			time.erase (i);
						
			TimeChanged ();
			break;
		}
	}
}

void
Selection::remove (jack_nframes_t start, jack_nframes_t end)
{
}

void
Selection::remove (AutomationLine *ac)
{
	list<AutomationLine*>::iterator i;
	if ((i = find (lines.begin(), lines.end(), ac)) != lines.end()) {
		lines.erase (i);
		LinesChanged();
	}
}

void
Selection::set (TimeAxisView* track)
{
	clear_tracks ();
	add (track);
}

void
Selection::set (AudioRegionView* r)
{
	clear_audio_regions ();
	add (r);
}

long
Selection::set (TimeAxisView* track, jack_nframes_t start, jack_nframes_t end)
{
	if ((start == 0 && end == 0) || end < start) {
		return 0;
	}

	if (time.empty()) {
		time.push_back (AudioRange (start, end, next_time_id++));
	} else {
		/* reuse the first entry, and remove all the rest */

		while (time.size() > 1) {
			time.pop_front();
		}
		time.front().start = start;
		time.front().end = end;
	}

	if (track) {
		time.track = track;
		time.group = track->edit_group();
	} else {
		time.track = 0;
		time.group = 0;
	}
	
	TimeChanged ();

	return time.front().id;
}

void
Selection::set (AutomationLine *ac)
{
	lines.clear();
	add (ac);
}

AudioRange&
TimeSelection::operator[] (unsigned long which)
{
	for (list<AudioRange>::iterator i = begin(); i != end(); ++i) {
		if ((*i).id == which) {
			return *i;
		}
	}
	fatal << compose (_("programming error: request for non-existent audio range (%1)!"), which) << endmsg;
	/*NOTREACHED*/
	return *(new AudioRange(0,0,0)); /* keep the compiler happy; never called */
}

jack_nframes_t 
TimeSelection::start ()
{
	if (empty()) {
		return 0;
	}

	jack_nframes_t first = max_frames;

	for (list<AudioRange>::iterator i = begin(); i != end(); ++i) {
		if ((*i).start < first) {
			first = (*i).start;
		}
	}
	return first;
}

jack_nframes_t 
TimeSelection::end_frame ()
{
	jack_nframes_t last = 0;

	/* XXX make this work like RegionSelection: no linear search needed */

	for (list<AudioRange>::iterator i = begin(); i != end(); ++i) {
		if ((*i).end > last) {
			last = (*i).end;
		}
	}
	return last;
}

jack_nframes_t
TimeSelection::length()
{
	return end_frame() - start() + 1;
}
	
