/*
    Copyright (C) 2003 Paul Davis 
    Author: Taybin Rutkin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: audio_library.cc,v 1.14 2003/10/30 03:41:37 trutkin Exp $
*/

#include <cstdio> // Needed so that libraptor (included in lrdf) won't complain
#include <iostream>
#include <sstream>
#include <cctype>

#include <lrdf.h>

#include <pbd/compose.h>

#include <ardour/ardour.h>
#include <ardour/configuration.h>
#include <ardour/audio_library.h>
#include <ardour/utils.h>
#include <ardour/version.h>

#include "i18n.h"

using namespace std;
using namespace ARDOUR;

namespace std {
struct UriSorter {
	bool operator() (string a, string b) const { 
		return cmp_nocase(Library->get_label(a), Library->get_label(b)) == -1; 
	}
}; 
};

static char* GROUP = "http://ardour.sf.net/ontology/Group";
static char* GROUP_ID = "http://ardour.sf.net/ontology/group_id";
static char* SOUNDFILE = "http://ardour.sf.net/ontology/Soundfile";
static char* SOUNDFILE_ID = "http://ardour.sf.net/ontology/soundfile_id";
static char* hasFile = "http://ardour.sf.net/ontology/hasFile";
static char* AUDIODB = "http://ardour.sf.net/ontology/AudioDB";

int touch_file(string path);

int
touch_file(string path)
{
	FILE* file = fopen(path.c_str(), "a");
	fclose(file);
	return 1;
}

AudioLibrary::AudioLibrary ()
{
	src = new string("file:" + Config->get_user_ardour_path() + "sfdb");

	// workaround for possible bug in raptor that crashes when saving to a
	// non-existant file.
	touch_file(Config->get_user_ardour_path() + "sfdb");

	lrdf_read_file(src->c_str());

	lrdf_statement pattern;

	pattern.subject = strdup(src->c_str());
	pattern.predicate = GROUP_ID;
	pattern.object = 0;
	pattern.object_type = lrdf_literal;

	lrdf_statement* matches = lrdf_matches(&pattern);

	// if empty DB, create basic schema
	if (matches == 0) {
		initialize_db ();
		group_id = 0;
		soundfile_id = 0;
	} else {
		group_id = atoi(matches->object);
		lrdf_free_statements(matches);
		
		pattern.predicate = SOUNDFILE_ID;
		matches = lrdf_matches(&pattern);
		
		if (matches) {
			soundfile_id = atoi(matches->object);
		} else {
			warning << _("Could not find starting point for SOUNDFILE_ID") << endmsg;
		}
	}

	lrdf_free_statements(matches);
	free(pattern.subject);
	save_changes();
}

AudioLibrary::~AudioLibrary ()
{
	save_groups();
	
	delete src;
}

void
AudioLibrary::initialize_db ()
{
	// define ardour:Group
	lrdf_add_triple(src->c_str(), GROUP, RDF_TYPE, RDFS_CLASS, lrdf_uri);
	// define ardour:Soundfile
	lrdf_add_triple(src->c_str(), SOUNDFILE, RDF_TYPE, RDFS_CLASS, lrdf_uri);
	// define ardour:AudioDB
	lrdf_add_triple(src->c_str(), AUDIODB, RDF_TYPE, RDFS_CLASS, lrdf_uri);

	// set file:sfdb to type AUDIODB
	lrdf_add_triple(src->c_str(), src->c_str(), RDF_TYPE, AUDIODB, lrdf_uri);

	// add intergral fields
	add_field("channels");
	add_field("samplerate");
	add_field("resolution");
	add_field("format");

	// define ardour:group_id
	lrdf_add_triple(src->c_str(), GROUP_ID, RDF_TYPE, 
			RDF_BASE "Property", lrdf_uri);
	lrdf_add_triple(src->c_str(), GROUP_ID, RDFS_BASE "range", 
			AUDIODB, lrdf_uri);
	lrdf_add_triple(src->c_str(), GROUP_ID, RDFS_BASE "domain", 
			RDF_BASE "Literal", lrdf_uri);

	// set AUDIODB group_id to 0
	lrdf_add_triple(src->c_str(), src->c_str(), GROUP_ID, 
			"0", lrdf_literal);

	// define ardour:soundfile_id
	lrdf_add_triple(src->c_str(), SOUNDFILE_ID, RDF_TYPE,
					RDF_BASE "Property", lrdf_uri);
	lrdf_add_triple(src->c_str(), SOUNDFILE_ID, RDFS_BASE "range",
					AUDIODB, lrdf_uri);
	lrdf_add_triple(src->c_str(), SOUNDFILE_ID, RDFS_BASE "domain",
					RDF_BASE "Literal", lrdf_uri);
	
	// set AUDIODB soundfile_id to 0
	lrdf_add_triple(src->c_str(), src->c_str(), SOUNDFILE_ID,
					"0", lrdf_literal);
}

void
AudioLibrary::save_changes ()
{
	lrdf_rebuild_caches();
	lrdf_export_by_source(src->c_str(), src->substr(5).c_str());
}

string
AudioLibrary::add_group (string group, string parent_uri)
{
	string local_group(compose("%1/group_id/%2", *src, ++group_id));

	lrdf_add_triple(src->c_str(), local_group.c_str(), 
					RDFS_BASE "label", group.c_str(), lrdf_literal);

	if (parent_uri == ""){
		lrdf_add_triple(src->c_str(), local_group.c_str(), 
						RDFS_SUBCLASSOF, GROUP, lrdf_uri);
	} else {
		lrdf_add_triple(src->c_str(), local_group.c_str(), 
						RDFS_SUBCLASSOF, parent_uri.c_str(), lrdf_uri);
	}

	EMIT_SIGNAL added_group(local_group, parent_uri);

	return local_group;
}

void
AudioLibrary::save_groups ()
{
	lrdf_statement pattern;
	pattern.subject = strdup(src->c_str());
	pattern.predicate = GROUP_ID;
	pattern.object = 0;
	pattern.object_type = lrdf_literal;
	lrdf_remove_matches(&pattern);
	free(pattern.subject);

	string id(compose("%1", group_id));
	lrdf_add_triple(src->c_str(), src->c_str(), GROUP_ID, 
			id.c_str(), lrdf_literal);

	save_changes ();
}

void
AudioLibrary::remove_group (string uri)
{
	lrdf_uris* ulist = lrdf_get_instances(uri.c_str());
	for (int i = 0; ulist && i < ulist->count; ++i){
		remove_member(ulist->items[i]);
	}
	
	ulist = lrdf_get_subclasses(uri.c_str());
	for (int i = 0; ulist && i < ulist->count; ++i){
		remove_group(ulist->items[i]);
	}

	lrdf_remove_uri_matches(uri.c_str());
	save_changes ();

	EMIT_SIGNAL removed_group(uri);
}

void
AudioLibrary::get_groups (list<string>& groups, string uri)
{
	lrdf_uris* ulist;

	if (uri == ""){
		ulist = lrdf_get_subclasses(GROUP);
	} else {
		ulist = lrdf_get_subclasses(uri.c_str());
	}

	for (int i = 0; ulist && i < ulist->count; ++i){
		groups.push_back(ulist->items[i]);
	}

	lrdf_free_uris(ulist);

	UriSorter cmp;
	groups.sort(cmp);
}

string
AudioLibrary::add_member (string member, string parent_uri)
{
	string local_member(compose("%1/soundfile_id/%2", *src, ++soundfile_id));
	string file_uri(compose("file:%1", member));

	lrdf_add_triple(src->c_str(), local_member.c_str(), RDF_TYPE, 
			SOUNDFILE, lrdf_uri);
	lrdf_add_triple(src->c_str(), local_member.c_str(), hasFile,
					file_uri.c_str(), lrdf_uri);

	string::size_type size = member.find_last_of('/');
	string label = member.substr(++size);

	lrdf_add_triple(src->c_str(), local_member.c_str(), RDFS_BASE "label", 
			label.c_str(), lrdf_literal);

	if (parent_uri == ""){
		lrdf_add_triple(src->c_str(), local_member.c_str(), RDF_TYPE,
				GROUP, lrdf_uri);
	} else {
		lrdf_add_triple(src->c_str(), local_member.c_str(), RDF_TYPE, 
			parent_uri.c_str(), lrdf_uri);
	}

	lrdf_statement pattern;
	pattern.subject = strdup(src->c_str());
	pattern.predicate = SOUNDFILE_ID;
	pattern.object = 0;
	pattern.object_type = lrdf_literal;
	lrdf_remove_matches(&pattern);
	free(pattern.subject);

	string id(compose("%1", soundfile_id));
	lrdf_add_triple(src->c_str(), src->c_str(), SOUNDFILE_ID, 
			id.c_str(), lrdf_literal);

	save_changes ();

	EMIT_SIGNAL added_member (local_member, parent_uri);

	return local_member;
}

void
AudioLibrary::remove_member (string uri)
{
	lrdf_remove_uri_matches (uri.c_str());

	save_changes ();

	EMIT_SIGNAL removed_member(uri);
}

void
AudioLibrary::get_members (list<string>& members, string parent_uri)
{
	lrdf_uris* ulist;

	if (parent_uri == ""){
		ulist = lrdf_get_instances(GROUP);
	} else {
		ulist = lrdf_get_instances(parent_uri.c_str());
	}

	for (int i = 0; ulist && i < ulist->count; ++i){
		members.push_back(ulist->items[i]);
	}

	lrdf_free_uris(ulist);

	UriSorter cmp;
	members.sort(cmp);
}

void
AudioLibrary::search_members_and (list<string>& members, 
								  const map<string,string>& fields)
{
	lrdf_statement **head;
	lrdf_statement* pattern = 0;
	lrdf_statement* old = 0;
	head = &pattern;

	map<string,string>::const_iterator i;
	for (i = fields.begin(); i != fields.end(); ++i){
		pattern = new lrdf_statement;
		pattern->subject = "?";
		pattern->predicate = strdup(field_uri(i->first).c_str());
		pattern->object = strdup((i->second).c_str());
		pattern->next = old;

		old = pattern;
	}

	if (*head != 0) {
		lrdf_uris* ulist = lrdf_match_multi(*head);
		for (int j = 0; ulist && j < ulist->count; ++j) {
//			printf("AND: %s\n", ulist->items[j]);
			members.push_back(ulist->items[j]);
		}
		lrdf_free_uris(ulist);

		UriSorter cmp;
		members.sort(cmp);
		members.unique();
	}

	// memory clean up
	pattern = *head;
	while(pattern){
		free(pattern->predicate);
		free(pattern->object);
		old = pattern;
		pattern = pattern->next;
		delete old;
	}
}

void
AudioLibrary::search_members_or (list<string>& members, 
								 const map<string,string>& fields)
{
	map<string,string>::const_iterator i;

	lrdf_statement pattern;
	for (i = fields.begin(); i != fields.end(); ++i) {
		pattern.subject = 0;
		pattern.predicate = strdup(field_uri(i->first).c_str());
		pattern.object = strdup((i->second).c_str());
		pattern.object_type = lrdf_literal;

		lrdf_statement* matched = lrdf_matches(&pattern);

		lrdf_statement* old = matched;
		while(matched) {
//			printf ("OR: %s\n", matched->subject);
			members.push_back(matched->subject);
			matched = matched->next;
		}

		free(pattern.predicate);
		free(pattern.object);
		lrdf_free_statements (old);
	}

	UriSorter cmp;
	members.sort(cmp);
	members.unique();
}

string
AudioLibrary::get_member_filename (string uri)
{
	lrdf_statement pattern;
	pattern.subject = strdup(uri.c_str());
	pattern.predicate = hasFile;
	pattern.object = 0;
	pattern.object_type = lrdf_uri;
	
	lrdf_statement* matches = lrdf_matches(&pattern);
	if (matches) {
		string file = matches->object;
		lrdf_free_statements(matches);

		string::size_type pos = file.find(":");
		return file.substr(++pos);
	} else {
		warning << _("Could not find member filename") << endmsg;
		return "-Unknown-";
	}
}

void
AudioLibrary::add_field (string name)
{
	string local_field = field_uri(name);
	lrdf_statement pattern;
	pattern.subject = strdup(local_field.c_str());
	pattern.predicate = RDF_TYPE;
	pattern.object = RDF_BASE "Property";
	pattern.object_type = lrdf_uri;

	if(lrdf_exists_match(&pattern)) {
		return;
	}

	// of type rdf:Property
	lrdf_add_triple(src->c_str(), local_field.c_str(), RDF_TYPE, 
			RDF_BASE "Property", lrdf_uri);
	// of range ardour:Soundfile
	lrdf_add_triple(src->c_str(), local_field.c_str(), RDFS_BASE "range",
			SOUNDFILE, lrdf_uri);
	// of domain rdf:Literal
	lrdf_add_triple(src->c_str(), local_field.c_str(), RDFS_BASE "domain", 
					RDF_BASE "Literal", lrdf_uri);

	set_label (local_field, name);
	
	save_changes();

	EMIT_SIGNAL fields_changed();
}

void
AudioLibrary::get_fields (list<string>& fields)
{
	lrdf_statement pattern;

	pattern.subject = 0;
	pattern.predicate = RDFS_BASE "range";
	pattern.object = SOUNDFILE;
	pattern.object_type = lrdf_uri;

	lrdf_statement* matches = lrdf_matches(&pattern);

	lrdf_statement* current = matches;
	lrdf_statement* old;
	while (current != 0) {
		fields.push_back(get_label(current->subject));

		old = current;
		current = current->next;
	}

	lrdf_free_statements(matches);

	fields.sort();
	fields.unique();
}

void
AudioLibrary::remove_field (string name)
{
	lrdf_remove_uri_matches(field_uri(name).c_str());
	save_changes();
	EMIT_SIGNAL fields_changed ();	
}

string 
AudioLibrary::get_field (string uri, string field)
{
	lrdf_statement pattern;

	pattern.subject = strdup(uri.c_str());

	pattern.predicate = strdup(field_uri(field).c_str());

	pattern.object = 0;
	pattern.object_type = lrdf_literal;

	lrdf_statement* matches = lrdf_matches(&pattern);
	free(pattern.subject);
	free(pattern.predicate);

	stringstream object;
	if (matches != 0){
		object << matches->object;
	}

	lrdf_free_statements(matches);
	return object.str();
}

void 
AudioLibrary::set_field (string uri, string field, string literal)
{
	lrdf_statement pattern;

	pattern.subject = strdup(uri.c_str());

	string local_field = field_uri(field);
	pattern.predicate = strdup(local_field.c_str());

	pattern.object = 0;
	pattern.object_type = lrdf_literal;

	lrdf_remove_matches(&pattern);
	free(pattern.subject);
	free(pattern.predicate);

	lrdf_add_triple(src->c_str(), uri.c_str(), local_field.c_str(), 
			literal.c_str(), lrdf_literal);

	save_changes();

	EMIT_SIGNAL fields_changed();
}

string
AudioLibrary::field_uri (string name)
{
	stringstream local_field;
	local_field << *src << "/fields/" << name;

	return local_field.str();
}

string
AudioLibrary::get_label (string uri)
{
	lrdf_statement pattern;
	pattern.subject = strdup(uri.c_str());
	pattern.predicate = RDFS_BASE "label";
	pattern.object = 0;
	pattern.object_type = lrdf_literal;

	lrdf_statement* matches = lrdf_matches (&pattern);
	free(pattern.subject);

	stringstream label;
	if (matches != 0){
		label << matches->object;
	}

	return label.str();
}

void
AudioLibrary::set_label (string uri, string label)
{
	lrdf_statement pattern;
	pattern.subject = strdup(uri.c_str());
	pattern.predicate = RDFS_BASE "label";
	pattern.object = 0;
	pattern.object_type = lrdf_literal;

	lrdf_remove_matches(&pattern);
	free(pattern.subject);

	lrdf_add_triple(src->c_str(), uri.c_str(), RDFS_BASE "label", 
			label.c_str(), lrdf_literal);
}
