/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: automation_event.cc,v 1.6 2004/01/15 01:56:11 pbd Exp $
*/

#include <algorithm>
#include <ardour/automation_event.h>

using namespace ARDOUR;

struct TimeSort {
    bool operator() (const AutomationEvent& a, const AutomationEvent& b) {
	    return a.when < b.when;
    }
};

AutomationList::AutomationList ()
{
	_frozen = false;
	changed_when_thawed = false;
}

void
AutomationList::maybe_signal_changed ()
{
	if (_frozen) {
		changed_when_thawed = true;
	} else {
		Changed();
	}
}

void
AutomationList::clear ()
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		_list.clear ();
	}

	maybe_signal_changed ();
}

void 
AutomationList::add (const AutomationEvent& ev)
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		iterator x;

		for (x = _list.begin(); x != _list.end(); ++x) {

			/* only one point allowed per time point */

			if ((*x).when == ev.when) {
				(*x).value = ev.value;
				break;
			} 

			if ((*x).when >= ev.when) {
				_list.insert (x, ev);
				break;
			}
		}

		if (x == _list.end()) {
			_list.push_back (ev);
		}
	}

	maybe_signal_changed ();
}

void
AutomationList::erase (AutomationList::iterator i)
{
	_list.erase (i);
	maybe_signal_changed ();
}

void
AutomationList::erase (AutomationList::iterator start, AutomationList::iterator end)
{
	_list.erase (start, end);
	maybe_signal_changed ();
}	

void
AutomationList::erase_range (jack_nframes_t start, jack_nframes_t endt)
{
	bool erased = false;
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		iterator x, tmp;
		
		for (x = _list.begin(); x != _list.end(); ) {
			
			tmp = x;
			++tmp;
			
			if ((*x).when >= endt) {
				break;
			}
			if ((*x).when >= start) {
				_list.erase (x);
				erased = true;
			}
			x = tmp;
		}
	}

	if (erased) {
		maybe_signal_changed ();
	}
}

void
AutomationList::adjust_range (iterator start, iterator end, float delta, float lower, float upper)
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		
		while (start != end) {
			(*start).value += delta;
			(*start).value = std::min ((*start).value, upper);
			(*start).value = std::max ((*start).value, lower);
			++start;
		}
	}

	maybe_signal_changed ();
}

void
AutomationList::modify (iterator iter, jack_nframes_t when, float val)
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		(*iter).when = when;
		(*iter).value = val;
	}
	
	maybe_signal_changed ();
}

AutomationList::iterator
AutomationList::nth (unsigned long n)
{
	iterator x;
	for (x = _list.begin(); x != _list.end() && n; --n, ++x);
	return x;
}		

std::pair<AutomationList::iterator,AutomationList::iterator>
AutomationList::control_points_adjacent (jack_nframes_t xval)
{
	LockMonitor lm (lock, __LINE__, __FILE__);
	iterator i;
	std::pair<iterator,iterator> ret;

	ret.first = _list.end();
	ret.second = _list.end();

	for (i = _list.begin(); i != _list.end(); ++i) {
		
		if (ret.first == _list.end()) {
			if ((*i).when >= xval) {
				if (i != _list.begin()) {
					ret.first = i;
					--ret.first;
				} else {
					return ret;
				}
			}
		} 
		
		if ((*i).when > xval) {
			ret.second = i;
			break;
		}
	}

	return ret;
}

void
AutomationList::freeze ()
{
	_frozen = true;
}

void
AutomationList::thaw ()
{
	_frozen = false;
	if (changed_when_thawed) {
		EMIT_SIGNAL Changed();
	}
}




