/*
    Copyright (C) 2003 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: crossfade.cc,v 1.4 2003/11/25 18:02:01 pbd Exp $
*/

#include <sigc++/bind.h>

#include <ardour/crossfade.h>
#include <ardour/crossfade_compare.h>
#include <ardour/audioregion.h>
#include <ardour/playlist.h>
#include <ardour/utils.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

/* XXX if and when we ever implement parallel processing of the process()
   callback, these will need to be handled on a per-thread basis.
*/

Sample* Crossfade::crossfade_buffer_left = 0;
Sample* Crossfade::crossfade_buffer_right = 0;

void
Crossfade::set_buffer_size (jack_nframes_t sz)
{
	if (crossfade_buffer_left) {
		delete [] crossfade_buffer_left;
		crossfade_buffer_left = 0;
	}

	if (crossfade_buffer_right) {
		delete [] crossfade_buffer_right;
		crossfade_buffer_right = 0;
	}

	if (sz) {
		crossfade_buffer_left = new Sample[sz];
		crossfade_buffer_right = new Sample[sz];
	}
}

bool
Crossfade::operator== (const Crossfade& other)
{
	return (_right == other._right) && (_left == other._left);
}

Crossfade::Crossfade (ARDOUR::AudioRegion& r, ARDOUR::AudioRegion& l, bool act)
	: _right (&r), 
	  _left (&l), 
	  _active (act),
	  _fade_in (0.0, 2.0, 1.0), // linear (gain coefficient) => -inf..+6dB
	  _fade_out (0.0, 2.0, 1.0) // linear (gain coefficient) => -inf..+6dB
{
	overlap_type = _right->coverage (_left->first_frame(), _left->last_frame());
	last_length = length();
	_current_state_id = 0;
	_in_update = false;

	_fade_out.add_control_point (0.0, 1.0, false);
	_fade_out.add_control_point (last_length * 0.1, 0.99, false);  
	_fade_out.add_control_point (last_length * 0.2, 0.97, false);
	_fade_out.add_control_point (last_length * 0.35, 0.8, false);
	_fade_out.add_control_point (last_length * 0.5, 0.5, false);
	_fade_out.add_control_point (last_length * 0.6, 0.3, false);
	_fade_out.add_control_point (last_length * 0.8, 0.03, false);
	_fade_out.add_control_point (last_length * 0.9, 0.01, false);
	_fade_out.add_control_point (last_length, 0.0, false);

	_fade_in.add_control_point (0.0, 0.0, false);
	_fade_in.add_control_point (last_length * 0.1,  0.01, false);
	_fade_in.add_control_point (last_length * 0.2,  0.03, false);
	_fade_in.add_control_point (last_length * 0.4,  0.3, false);
	_fade_in.add_control_point (last_length * 0.5,  0.5, false);
	_fade_in.add_control_point (last_length * 0.65, 0.8, false);
	_fade_in.add_control_point (last_length * 0.8,  0.97, false);
	_fade_in.add_control_point (last_length * 0.9,  0.99, false);
	_fade_in.add_control_point (last_length, 1.0, false);

	_fade_in.Changed.connect (slot (*this, &Crossfade::fade_in_changed));
	_fade_out.Changed.connect (slot (*this, &Crossfade::fade_out_changed));

	save_state ("initial");
}

Crossfade::Crossfade (const Playlist& playlist, XMLNode& node)
	:  _fade_in (0.0, 2.0, 1.0), // linear (gain coefficient) => -inf..+6dB
	   _fade_out (0.0, 2.0, 1.0) // linear (gain coefficient) => -inf..+6dB
{
	Region* r;
	XMLProperty* prop;
	id_t id;

	if ((prop = node.property ("right")) == 0) {
		error << _("Crossfade: no \"right\" region in state") << endmsg;
		throw failed_constructor();
	}
	
	sscanf (prop->value().c_str(), "%Lu", &id);

	if ((r = playlist.find_region (id)) == 0) {
		error << compose (_("Crossfade: no right region %1 found in playlist %2"), id, playlist.name())
		      << endmsg;
		throw failed_constructor();
	}
	
	if ((_right = dynamic_cast<AudioRegion*> (r)) == 0) {
		throw failed_constructor();
	}

	if ((prop = node.property ("left")) == 0) {
		error << _("Crossfade: no \"left\" region in state") << endmsg;
		throw failed_constructor();
	}

	sscanf (prop->value().c_str(), "%Lu", &id);

	if ((r = playlist.find_region (id)) == 0) {
		error << compose (_("Crossfade: no left region %1 found in playlist %2"), id, playlist.name())
		      << endmsg;
		throw failed_constructor();
	}

	if ((_left = dynamic_cast<AudioRegion*> (r)) == 0) {
		throw failed_constructor();
	}

	overlap_type = _right->coverage (_left->first_frame(), _left->last_frame());
	_active = true;
	last_length = length();
	_current_state_id = 0;
	_in_update = false;

	set_state (node);

	_fade_in.Changed.connect (slot (*this, &Crossfade::fade_in_changed));
	_fade_out.Changed.connect (slot (*this, &Crossfade::fade_out_changed));

	save_state ("initial");
}

Crossfade::~Crossfade ()
{
	for (vector<State*>::iterator i = states.begin(); i != states.end(); ++i) {
		delete *i;
	}
}

jack_nframes_t
Crossfade::length () const
{
	return _left->last_frame() - _right->first_frame();
}

void
Crossfade::update ()
{
	jack_nframes_t newlen = length();

	if (newlen != last_length) {

		/* move the terminal point to the new end of the xfade */
		
		_in_update = true;
		_fade_in.reset_control_point (_fade_in.n_control_points(), 1.0, newlen);
		_fade_out.reset_control_point (_fade_out.n_control_points(), 0.0, newlen);
		_in_update = false;

		last_length = newlen;
		
		save_state ("updated");
		
		EMIT_SIGNAL Changed();
	}
}

jack_nframes_t 
Crossfade::read_at (Sample *buf, Sample *mixdown_buffer, 
		    float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n)
{
	jack_nframes_t offset;
	jack_nframes_t to_write;
	jack_nframes_t first_frame;

	if (!_active) {
		return 0;
	}

	first_frame = _right->position();

	if (position < first_frame) {

		/* handle an initial section of the read area that we do not
		   cover.
		*/
		

		offset = first_frame - position;

		if (offset < cnt) {
			cnt -= offset;
		} else {
			return 0;
		}

		position = first_frame;
		buf += offset;
	}

	_left->read_at (crossfade_buffer_left, mixdown_buffer, gain_buffer, position, cnt, chan_n);
	_right->read_at (crossfade_buffer_right, mixdown_buffer, gain_buffer, position, cnt, chan_n);
	
	offset = position - first_frame;
	to_write = min (last_length - offset, cnt);

	float* fiv = new float[to_write];
	float* fov = new float[to_write];

	_fade_in.get_vector (offset, offset+to_write, fiv, to_write);
	_fade_out.get_vector (offset, offset+to_write, fov, to_write);

	/* note: although we have not explicitly taken into account the return values
	   from _left->read_at() or _right->read_at(), the length() function does this
	   implicitly. why? because it computes a value based on the left+right regions'
	   position and length, and so we know precisely how much data they could return. 
	*/
	
	for (jack_nframes_t n = 0; n < to_write; ++n) {
		buf[n] = (crossfade_buffer_left[n] * fov[n]) + (crossfade_buffer_right[n] * fiv[n]);
	}

	delete [] fov;
	delete [] fiv;

	return to_write;
}	

OverlapType 
Crossfade::coverage (jack_nframes_t start, jack_nframes_t end) const
{
	jack_nframes_t rstart = _right->position();
	jack_nframes_t rend = _left->last_frame();
	
	if ((start >= rstart) && (end <= rend)) {
		return OverlapInternal;
	}
	if ((end >= rstart) && (end <= rend)) {
		return OverlapStart;
	}
	if ((start >= rstart) && (start <= rend)) {
		return OverlapEnd;
	}
	if ((rstart >= start) && (rstart <= end) && (rend <= end)) {
		return OverlapExternal;
	}
	return OverlapNone;
}

void
Crossfade::refresh ()
{
	OverlapType ot;

	ot = _right->coverage (_left->first_frame(), _left->last_frame());

	switch (ot) {
	case OverlapNone:
	case OverlapInternal:
	case OverlapExternal:
		Invalidated (this);
		return;

	default:
		break;
	}

	/* a bounds change may have caused us to switch to a new kind of overlap */

	if (ot != overlap_type) {
		Invalidated (this);
	} else {
		update ();
	}
}

void
Crossfade::set_active (bool yn)
{
	if (_active != yn) {
		_active = yn;
		save_state (_("active changed"));
		ActiveChanged ();
	}
}

void
Crossfade::fade_in_changed ()
{
	if (!_in_update) {
		update ();
	}
}

void
Crossfade::fade_out_changed ()
{
	if (!_in_update) {
		update ();
	}
}

Change
Crossfade::restore_state (StateManager::State& state)
{
	CrossfadeState* xfstate = dynamic_cast<CrossfadeState*> (&state);

	_active = xfstate->active;
	xfstate->fade_in_memento ();
	xfstate->fade_out_memento ();
	
	return Change (0);
}

StateManager::State*
Crossfade::state_factory (std::string why) const
{
	CrossfadeState* state = new CrossfadeState (why);
	
	state->fade_in_memento = _fade_in.get_memento ();
	state->fade_out_memento = _fade_out.get_memento ();
	state->active = _active;

	return state;
}
	
UndoAction
Crossfade::get_memento() const
{
	return bind (slot (*(const_cast<Crossfade *> (this)), &StateManager::use_state), _current_state_id);
}

XMLNode&
Crossfade::get_state () 
{
	XMLNode* node = new XMLNode (X_("Crossfade"));
	XMLNode* child;
	char buf[64];
	size_t limit;

	snprintf (buf, sizeof(buf), "%Lu", _left->id());
	node->add_property ("left", buf);
	snprintf (buf, sizeof(buf), "%Lu", _right->id());
	node->add_property ("right", buf);
	node->add_property ("active", (_active ? "yes" : "no"));

	child = node->add_child ("FadeIn");
	limit = _fade_in.n_control_points();

	for (size_t n = 0; n < limit; ++n) {
		double x, y;
		XMLNode* pnode = new XMLNode ("point");

		_fade_in.nth (n, x, y);

		snprintf (buf, sizeof (buf), "%g", x);
		pnode->add_property ("x", buf);
		snprintf (buf, sizeof (buf), "%g", y);
		pnode->add_property ("y", buf);
		child->add_child_nocopy (*pnode);
	}

	child = node->add_child ("FadeOut");
	limit = _fade_out.n_control_points();

	for (size_t n = 0; n < limit; ++n) {
		double x, y;
		char buf[64];
		XMLNode* pnode = new XMLNode ("point");

		_fade_out.nth (n, x, y);
		
		snprintf (buf, sizeof (buf), "%g", x);
		pnode->add_property ("x", buf);
		snprintf (buf, sizeof (buf), "%g", y);
		pnode->add_property ("y", buf);
		child->add_child_nocopy (*pnode);
	}

	return *node;
}

int
Crossfade::set_state (const XMLNode& node)
{
	XMLNodeConstIterator i;
	XMLNodeList children;
	XMLNode* fi;
	XMLNode* fo;
	const XMLProperty* prop;

	if ((fi = find_named_node (node, "FadeIn")) == 0) {
		return -1;
	}
	
	if ((fo = find_named_node (node, "FadeOut")) == 0) {
		return -1;
	}

	if ((prop = node.property ("active")) == 0) {
		_active = false;
	} else {
		set_active (prop->value() == "yes");
	}
	
	/* fade in */
	
	_fade_in.clear ();
	
	children = fi->children();
	
	for (i = children.begin(); i != children.end(); ++i) {
		if ((*i)->name() == "point") {
			float x;
			float y;
			
			prop = (*i)->property ("x");
			sscanf (prop->value().c_str(), "%f", &x);
			
			prop = (*i)->property ("y");
			sscanf (prop->value().c_str(), "%f", &y);
			
			_fade_in.add_control_point (x, y);
		}
	}
	
        /* fade out */
	
	_fade_out.clear ();

	children = fo->children();
	
	for (i = children.begin(); i != children.end(); ++i) {
		if ((*i)->name() == "point") {
			float x;
			float y;
			XMLProperty* prop;

			prop = (*i)->property ("x");
			sscanf (prop->value().c_str(), "%f", &x);

			prop = (*i)->property ("y");
			sscanf (prop->value().c_str(), "%f", &y);
			
			_fade_out.add_control_point (x, y);
		}
	}
		
	return 0;
}

