/*
    Copyright (C) 1999-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session_export.cc,v 1.16 2004/01/15 01:56:11 pbd Exp $
*/

/* see dither.cc for why we have to do this */

#define	_ISOC9X_SOURCE	1
#define _ISOC99_SOURCE	1
#include <math.h>
#undef  _ISOC99_SOURCE
#undef  _ISOC9X_SOURCE
#undef  __USE_SVID 
#define __USE_SVID 1
#include <stdlib.h>
#undef  __USE_SVID

#include <cmath>
#include <cstdlib>
#include <unistd.h>
#include <stdint.h>

#include <sigc++/bind.h>

#include <pbd/error.h>
#include <pbd/lockmonitor.h>

#include <ardour/dither.h>
#include <ardour/timestamps.h>
#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/export.h>
#include <ardour/sndfile_helpers.h>
#include <ardour/port.h>
#include <ardour/audioengine.h>
#include <ardour/diskstream.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

AudioExportSpecification::AudioExportSpecification ()
{
	init ();
}

AudioExportSpecification::~AudioExportSpecification ()
{
	clear ();
}

void
AudioExportSpecification::init ()
{
	dither = 0;
	src_state = 0;
	pos = 0;
	start_frame = 0;
	end_frame = 0;
	total_frames = 0;
	dataF = 0;
	dataF2 = 0;
	leftoverF = 0;
	max_leftover_frames = 0;
	leftover_frames = 0;
	output_data = 0;
	out_samples_max = 0;
	out = 0;
	data_width = 0;
	running = false;
	stop = false;
	progress = 0.0;
	status = 0;
	do_freewheel = true;
}

void
AudioExportSpecification::clear ()
{
	if (out) {
		sf_close (out);
	}

	dither_free (dither);

	if (output_data) {
		free (output_data);
	}
	if (dataF) {
		delete [] dataF;
	}
	if (src_state) {
		src_delete (src_state);
	}
	if (dataF2) {
		delete [] dataF2;
	}
	if (leftoverF) {
		delete [] leftoverF;
	}

	freewheel_connection.disconnect ();

	init ();
}

static int
convert_spec_to_info (AudioExportSpecification& spec, SF_INFO& sfinfo)
{
	if (spec.path.length() == 0) {
		error << _("Export: no output file specified") << endmsg;
		return -1;
	}

	/* XXX add checks that the directory path exists, and also 
	   check if we are overwriting an existing file...
	*/

	sfinfo.format = spec.format;
	sfinfo.samplerate = spec.sample_rate;
	sfinfo.frames = spec.end_frame - spec.start_frame + 1;
	sfinfo.channels = min (spec.channels, 2UL);

	return 0;
}

int
Session::start_audio_export (AudioExportSpecification& spec)
{
	if (prepare_to_export (spec)) {
		spec.running = false;
		spec.status = -1;
		return -1;
	}

	if (spec.do_freewheel) {
		spec.freewheel_connection = _engine.Freewheel.connect (bind (slot (*this, &Session::process_export), &spec));
		return _engine.freewheel (true);
	}
	else {
		spec.running = false;
		spec.status = -1;
		return -1;
	}
}

int
Session::stop_audio_export (AudioExportSpecification& spec)
{
	int ret;

	spec.stop = true;

	/* we are still freewheeling until we return from this ... */

	ret = _engine.freewheel (false);

	spec.clear ();

	stop_transport (false);

	return ret;
}

int 
Session::prepare_to_export (AudioExportSpecification& spec)
{
	int ret = -1;
	char errbuf[256];

	if (spec.start_frame >= spec.end_frame) {
		error << _("illegal frame range in export specification") << endmsg;
		return -1;
	}

	if ((spec.data_width = sndfile_data_width(spec.format)) == 0) {
		error << _("Bad data width size.  Report me!") << endmsg;
		return -1;
	}

	if (convert_spec_to_info (spec, spec.sfinfo)) {
		return -1;
	}

	/* XXX make sure we have enough disk space for the output */
	
	if ((spec.out = sf_open (spec.path.c_str(), SFM_WRITE, &spec.sfinfo)) == 0) {
		sf_error_str (0, errbuf, sizeof (errbuf) - 1);
		error << compose(_("Export: cannot open output file \"%1\" (%2)"), spec.path, errbuf) << endmsg;
		goto out;
	}

	spec.dataF = new float[current_block_size * spec.channels];
	spec.end_frame = min (current_end_frame(), spec.end_frame);
	spec.total_frames = spec.end_frame - spec.start_frame;

	if (spec.sample_rate != frame_rate()) {
		int err;

		if ((spec.src_state = src_new (spec.src_quality, spec.channels, &err)) == 0) {
			error << compose (_("cannot initialize sample rate conversion: %1"), src_strerror (err)) << endmsg;
			goto out;
		}
		
		spec.src_data.src_ratio = spec.sample_rate / (double) frame_rate();
		spec.out_samples_max = (jack_nframes_t) ceil (current_block_size * spec.src_data.src_ratio * spec.channels);
		spec.dataF2 = new float[spec.out_samples_max];

		spec.max_leftover_frames = 4 * current_block_size;
		spec.leftoverF = new float[spec.max_leftover_frames * spec.channels];
		spec.leftover_frames = 0;
		
	} else {
		spec.out_samples_max = current_block_size * spec.channels;
	}

	spec.dither = dither_new (spec.dither_type, spec.channels, spec.data_width);

	/* allocate buffers where dithering and output will occur */

	switch (spec.data_width) {
	case 8:
		spec.sample_bytes = sizeof(int8_t);
		break;

	case 16:
		spec.sample_bytes = sizeof(int16_t);
		break;

	case 24:
	case 32:
		spec.sample_bytes = sizeof(int32_t);
		break;

	default:
		spec.sample_bytes = 0; // float format
		break;
	}

	if (spec.sample_bytes) {
		spec.output_data = (void*) malloc (spec.sample_bytes * spec.out_samples_max);
	}

	
	wait_till_butler_finished ();

	/* get everyone to the right position */

	{
		LockMonitor (diskstream_lock, __LINE__, __FILE__);
		for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
			if ((*i)-> seek (spec.start_frame, true)) {
				error << compose (_("%1: cannot seek to %2 for export"),
						  (*i)->name(), spec.start_frame)
				      << endmsg;
				goto out;
			}
		}
	}

	/* make sure we are actually rolling */

	if (get_record_enabled()) {
		disable_record ();
	}
	
	set_transport_speed (1.0, false);
	butler_transport_work ();
	atomic_set (&butler_should_do_transport_work, 0);
	post_transport ();

	/* set up a few remaining details, and we are ready to go ... */

	spec.pos = spec.start_frame;
	ret = 0;

  out:
	return ret;
}

int
Session::process_export (jack_nframes_t nframes, AudioExportSpecification* spec)
{
	float* float_buffer = 0;
	unsigned long chn;
	unsigned long x;
	unsigned long i;
	sf_count_t written;
	int ret = -1;
	jack_nframes_t this_nframes;
	jack_nframes_t to_write = 0;
	char errbuf[256];

	if (!spec->running || spec->stop || (this_nframes = min ((spec->end_frame - spec->pos), nframes)) == 0) {
		/* cerr << "we have reached the end, with stop = " << spec->stop
		   << endl; */
		process_without_events (nframes);
		spec->running = false;
		return 0;
	}

	/* make sure we've caught up with disk i/o, since
	   we're running faster than realtime c/o JACK.
	*/

	wait_till_butler_finished ();
	
	/* do the usual stuff */
	
	process_without_events (nframes);

	/* and now export the results */

	nframes = this_nframes;

	memset (spec->dataF, 0, sizeof (spec->dataF[0]) * nframes * spec->channels);


	/* foreach output channel ... */
	
	for (chn = 0; chn < spec->channels; ++chn) {
		
		AudioExportPortMap::iterator mi = spec->port_map.find (chn);
		
		if (mi == spec->port_map.end()) {
			/* no ports exported to this channel */
			continue;
		}
		
		vector<PortChannelPair>& mapped_ports ((*mi).second);
		
		for (vector<PortChannelPair>::iterator t = mapped_ports.begin(); t != mapped_ports.end(); ++t) {
			
			/* OK, this port's output is supposed to appear on this channel 
			 */

			Port* port = (*t).first;
			Sample* port_buffer = port->get_buffer (nframes);

			/* now interleave the data from the channel into the float buffer */
				
			for (x = 0; x < nframes; ++x) {
				spec->dataF[chn+(x*spec->channels)] += (float) port_buffer[x];
			}
		}

		/* now do sample rate conversion */
		
		if (chn == (spec->channels-1) && spec->sample_rate != frame_rate()) {
			
			int err;
			
			spec->src_data.data_in = spec->dataF;
			spec->src_data.data_out = spec->dataF2;
			spec->src_data.input_frames = nframes;
			spec->src_data.output_frames = spec->out_samples_max / spec->channels;
			spec->src_data.end_of_input = ((spec->pos + nframes) >= spec->end_frame);

			if (spec->leftover_frames > 0) {
				memcpy (spec->leftoverF + (spec->leftover_frames * spec->channels), spec->dataF, nframes * spec->channels * sizeof(float));
				spec->src_data.data_in = spec->leftoverF;
				spec->src_data.input_frames += spec->leftover_frames;
			}
			
			if ((err = src_process (spec->src_state, &spec->src_data)) != 0) {
				error << compose (_("an error occured during sample rate conversion: %1"),
						  src_strerror (err))
				      << endmsg;
				goto out;
			}
							
			to_write = spec->src_data.output_frames_gen;

			spec->leftover_frames = spec->src_data.input_frames - spec->src_data.input_frames_used;

			if (spec->leftover_frames > 0) {
				if (spec->leftover_frames > spec->max_leftover_frames) {
					error << _("warning, leftover frames overflowed, glitches might occur in output") << endmsg;
					spec->leftover_frames = spec->max_leftover_frames;
				}
				memmove(spec->leftoverF, (char *) (spec->src_data.data_in + (spec->src_data.input_frames_used * spec->channels)),
				       spec->leftover_frames * spec->channels * sizeof(float));
			}
				
			float_buffer = spec->dataF2;

		} else {
			
			to_write = nframes;
			float_buffer = spec->dataF;
		}
	}

	if (spec->output_data) {
		memset (spec->output_data, 0, spec->sample_bytes * to_write * spec->channels);
	}

	switch (spec->data_width) {
	case 8:
	case 16:
	case 24:
		for (chn = 0; chn < spec->channels; ++chn) { 
			dither_run (spec->dither, chn, to_write, float_buffer, spec->output_data);
		}
		break;
		
	case 32:
		for (chn = 0; chn < spec->channels; ++chn) {
			
			int *ob = (int *) spec->output_data;
			
			for (x = 0; x < to_write; ++x) {
				i = chn + (x * spec->channels);
				
				if (float_buffer[i] > 1.0f) {
					ob[i] = 1;
				} else if (float_buffer[i] < -1.0f) {
					ob[i] = -1;
				} else {
					ob[i] = lrintf (float_buffer[i]);
				}
			}
		}
		break;
		
	default:
		for (x = 0; x < to_write * spec->channels; ++x) {
			if (float_buffer[x] > 1.0f) {
				float_buffer[x] = 1.0f;
			} else if (float_buffer[x] < -1.0f) {
				float_buffer[x] = -1.0f;
			}
		}
		break;
	}

	/* and export to disk */

	switch (spec->data_width) {
	case 8:
		/* XXXX no way to deliver 8 bit audio to libsndfile */
		written = to_write;
		break;

	case 16:
		written = sf_writef_short (spec->out, (short*) spec->output_data, to_write);
		break;

	case 24:
	case 32:
		written = sf_writef_int (spec->out, (int*) spec->output_data, to_write);
		break;

	default:
		written = sf_writef_float (spec->out, float_buffer, to_write);
		break;
	}
	
	if ((jack_nframes_t) written != to_write) {
		sf_error_str (spec->out, errbuf, sizeof (errbuf) - 1);
		error << compose(_("Export: could not write data to output file (%1)"), errbuf) << endmsg;
		goto out;
	}
	
	spec->pos += nframes;
	spec->progress = 1.0 - (((float) spec->end_frame - spec->pos) / spec->total_frames);

	/*
	cerr << "export position is now " << spec->pos << " =>" 
	     << spec->progress << "%" << endl;
	*/
	     
	/* and we're good to go */

	ret = 0;

  out: 
	if (ret) {
		sf_close (spec->out);
		spec->out = 0;
		unlink (spec->path.c_str());
		spec->running = false;
		spec->status = ret;
	}

	return ret;
}
